/*
* Copyright (c) 2024, Alliance for Open Media. All rights reserved
*
* This source code is subject to the terms of the BSD 2 Clause License and
* the Alliance for Open Media Patent License 1.0. If the BSD 2 Clause License
* was not distributed with this source code in the LICENSE file, you can
* obtain it at www.aomedia.org/license/software. If the Alliance for Open
* Media Patent License 1.0 was not distributed with this source code in the
* PATENTS file, you can obtain it at www.aomedia.org/license/patent.
*/

#ifndef AOM_OBMC_CONSTANTS_NEON_H_
#define AOM_OBMC_CONSTANTS_NEON_H_

#include "definitions.h"

extern const uint8_t obmc_variance_permute_idx[];

#endif // AOM_OBMC_CONSTANTS_NEON_H_
